@@ -55,7 +55,7 @@ read_long(const char *str, size_t len) {
 
 static VALUE
 calc_hash_key(ParseInfo pi, Val kval, char k1) {
-    VALUE	rkey;
+    volatile VALUE	rkey;
 
     if (':' == k1) {
 	rkey = rb_str_new(kval->key + 1, kval->klen - 1);
@@ -346,7 +346,11 @@ hat_num(ParseInfo pi, Val parent, Val kval, NumInfo ni) {
 		    args[2] = LONG2NUM(st->tm_mday);
 		    args[3] = LONG2NUM(st->tm_hour);
 		    args[4] = LONG2NUM(st->tm_min);
+#if NO_TIME_ROUND_PAD
+		    args[5] = rb_float_new((double)st->tm_sec + ((double)nsec) / 1000000000.0);
+#else
 		    args[5] = rb_float_new((double)st->tm_sec + ((double)nsec + 0.5) / 1000000000.0);
+#endif
 		    args[6] = LONG2NUM(ni->exp);
 		    parent->val = rb_funcall2(rb_cTime, oj_new_id, 7, args);
 #endif
@@ -384,9 +388,10 @@ hat_value(ParseInfo pi, Val parent, const char *key, size_t klen, volatile VALUE
 	int	len = (int)RARRAY_LEN(value);
 
 	if (2 == klen && 'u' == key[1]) {
-            VALUE	sc;
-	    VALUE	e1;
-
+            volatile VALUE	sc;
+	    volatile VALUE	e1;
+	    int			slen;
+	    
 	    if (0 == len) {
 		oj_set_error_at(pi, oj_parse_error_class, __FILE__, __LINE__, "Invalid struct data");
 		return 1;
@@ -394,9 +399,9 @@ hat_value(ParseInfo pi, Val parent, const char *key, size_t klen, volatile VALUE
 	    e1 = *RARRAY_PTR(value);
 	    // check for anonymous Struct
 	    if (T_ARRAY == rb_type(e1)) {
-		VALUE	args[1024];
-		VALUE	rstr;
-		int	i, cnt = (int)RARRAY_LEN(e1);
+		VALUE		args[1024];
+		volatile VALUE	rstr;
+		int		i, cnt = (int)RARRAY_LEN(e1);
 
 		for (i = 0; i < cnt; i++) {
 		    rstr = rb_ary_entry(e1, i);
@@ -411,27 +416,24 @@ hat_value(ParseInfo pi, Val parent, const char *key, size_t klen, volatile VALUE
             parent->val = rb_obj_alloc(sc);
             // If the JSON array has more entries than the struct class allows, we record an error.
 #ifdef RSTRUCT_LEN
+#if UNIFY_FIXNUM_AND_BIGNUM
+	    slen = (int)NUM2LONG(RSTRUCT_LEN(parent->val));
+#else // UNIFY_FIXNUM_AND_INTEGER
+	    slen = (int)RSTRUCT_LEN(parent->val);
+#endif // UNIFY_FIXNUM_AND_INTEGER
+#else
+	    slen = FIX2INT(rb_funcall2(parent->val, oj_length_id, 0, 0));
+#endif
             // MRI >= 1.9
-            if (len - 1 > RSTRUCT_LEN(parent->val)) {
+            if (len - 1 > slen) {
 		oj_set_error_at(pi, oj_parse_error_class, __FILE__, __LINE__, "Invalid struct data");
             } else {
-                MEMCPY(RSTRUCT_PTR(parent->val), RARRAY_PTR(value) + 1, VALUE, len - 1);
-            }
-#else
-            {
-		// MRI < 1.9 or Rubinius
-		int	slen = FIX2INT(rb_funcall2(parent->val, oj_length_id, 0, 0));
 		int	i;
 
-		if (len - 1 > slen) {
-		    oj_set_error_at(pi, oj_parse_error_class, __FILE__, __LINE__, "Invalid struct data");
-		} else {
-		    for (i = 0; i < slen; i++) {
-			rb_struct_aset(parent->val, INT2FIX(i), RARRAY_PTR(value)[i + 1]);
-		    }
+		for (i = 0; i < slen; i++) {
+		    rb_struct_aset(parent->val, INT2FIX(i), RARRAY_PTR(value)[i + 1]);
 		}
             }
-#endif
 	    return 1;
 	} else if (3 <= klen && '#' == key[1]) {
 	    volatile VALUE	*a;
@@ -476,8 +478,8 @@ set_obj_ivar(Val parent, Val kval, VALUE value) {
 
     if ('~' == *key && Qtrue == rb_obj_is_kind_of(parent->val, rb_eException)) {
 	if (5 == klen && 0 == strncmp("~mesg", key, klen)) {
-	    VALUE	args[1];
-	    VALUE	prev = parent->val;
+	    VALUE		args[1];
+	    volatile VALUE	prev = parent->val;
 
 	    args[0] = value;
 	    parent->val = rb_class_new_instance(1, args, rb_class_of(parent->val));
@@ -648,8 +650,8 @@ hash_set_value(ParseInfo pi, Val kval, VALUE value) {
 	    }
 	} else {
 	    if (3 <= klen && '^' == *key && '#' == key[1] && T_ARRAY == rb_type(value)) {
-		long	len = RARRAY_LEN(value);
-		VALUE	*a = RARRAY_PTR(value);
+		long		len = RARRAY_LEN(value);
+		volatile VALUE	*a = RARRAY_PTR(value);
 	
 		if (2 != len) {
 		    oj_set_error_at(pi, oj_parse_error_class, __FILE__, __LINE__, "invalid hash pair");
@@ -672,6 +674,7 @@ hash_set_value(ParseInfo pi, Val kval, VALUE value) {
     case T_OBJECT:
 	set_obj_ivar(parent, kval, value);
 	break;
+    case T_MODULE:
     case T_CLASS:
 	if (0 == parent->odd_args) {
 	    oj_set_error_at(pi, oj_parse_error_class, __FILE__, __LINE__, "%s is not an odd class", rb_class2name(rb_obj_class(parent->val)));
@@ -771,6 +774,7 @@ oj_object_parse(int argc, VALUE *argv, VALUE self) {
 
     pi.options = oj_default_options;
     pi.handler = Qnil;
+    pi.err_class = Qnil;
     oj_set_object_callbacks(&pi);
 
     if (T_STRING == rb_type(*argv)) {
@@ -786,6 +790,7 @@ oj_object_parse_cstr(int argc, VALUE *argv, char *json, size_t len) {
 
     pi.options = oj_default_options;
     pi.handler = Qnil;
+    pi.err_class = Qnil;
     oj_set_strict_callbacks(&pi);
     pi.end_hash = end_hash;
     pi.start_hash = start_hash;
