@@ -1,21 +1,21 @@
 /* parse.c
  * Copyright (c) 2013, Peter Ohler
  * All rights reserved.
- * 
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are met:
- * 
+ *
  *  - Redistributions of source code must retain the above copyright notice, this
  *    list of conditions and the following disclaimer.
- * 
+ *
  *  - Redistributions in binary form must reproduce the above copyright notice,
  *    this list of conditions and the following disclaimer in the documentation
  *    and/or other materials provided with the distribution.
- * 
+ *
  *  - Neither the name of Peter Ohler nor the names of its contributors may be
  *    used to endorse or promote products derived from this software without
  *    specific prior written permission.
- * 
+ *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
@@ -42,11 +42,6 @@
 // Workaround in case INFINITY is not defined in math.h or if the OS is CentOS
 #define OJ_INFINITY	(1.0/0.0)
 
-#ifdef RUBINIUS_RUBY
-#define NUM_MAX		0x07FFFFFF
-#else
-#define NUM_MAX		(FIXNUM_MAX >> 8)
-#endif
 //#define EXP_MAX		1023
 #define EXP_MAX		100000
 #define DEC_MAX		15
@@ -255,6 +250,11 @@ read_escaped_str(ParseInfo pi, const char *start) {
 
 		    s++;
 		    if ('\\' != *s || 'u' != *(s + 1)) {
+			if (Yes == pi->options.allow_invalid) {
+			    s--;
+			    unicode_to_chars(pi, &buf, code);
+			    break;
+			}
 			pi->cur = s;
 			oj_set_error_at(pi, oj_parse_error_class, __FILE__, __LINE__, "invalid escaped character");
 			buf_cleanup(&buf);
@@ -389,15 +389,14 @@ static void
 read_num(ParseInfo pi) {
     struct _NumInfo	ni;
     Val			parent = stack_peek(&pi->stack);
-    int			zero_cnt = 0;
 
     ni.str = pi->cur;
     ni.i = 0;
     ni.num = 0;
     ni.div = 1;
+    ni.di = 0;
     ni.len = 0;
     ni.exp = 0;
-    ni.dec_cnt = 0;
     ni.big = 0;
     ni.infinity = 0;
     ni.nan = 0;
@@ -426,21 +425,17 @@ read_num(ParseInfo pi) {
 	pi->cur += 3;
 	ni.nan = 1;
     } else {
+	int	dec_cnt = 0;
+
 	for (; '0' <= *pi->cur && *pi->cur <= '9'; pi->cur++) {
-	    ni.dec_cnt++;
-	    if (ni.big) {
-		ni.big++;
-	    } else {
+	    if (0 < ni.i) {
+		dec_cnt++;
+	    }
+	    if (!ni.big) {
 		int	d = (*pi->cur - '0');
 
-		if (0 == d) {
-		    zero_cnt++;
-		} else {
-		    zero_cnt = 0;
-		}
-		// TBD move size check here
 		ni.i = ni.i * 10 + d;
-		if (LONG_MAX <= ni.i || DEC_MAX < ni.dec_cnt - zero_cnt) {
+		if (INT64_MAX <= ni.i || DEC_MAX < dec_cnt) {
 		    ni.big = 1;
 		}
 	    }
@@ -450,16 +445,13 @@ read_num(ParseInfo pi) {
 	    for (; '0' <= *pi->cur && *pi->cur <= '9'; pi->cur++) {
 		int	d = (*pi->cur - '0');
 
-		if (0 == d) {
-		    zero_cnt++;
-		} else {
-		    zero_cnt = 0;
+		if (0 < ni.num || 0 < ni.i) {
+		    dec_cnt++;
 		}
-		ni.dec_cnt++;
-		// TBD move size check here
 		ni.num = ni.num * 10 + d;
 		ni.div *= 10;
-		if (LONG_MAX <= ni.div || DEC_MAX < ni.dec_cnt - zero_cnt) {
+		ni.di++;
+		if (INT64_MAX <= ni.div || DEC_MAX < dec_cnt) {
 		    ni.big = 1;
 		}
 	    }
@@ -485,9 +477,19 @@ read_num(ParseInfo pi) {
 		ni.exp = -ni.exp;
 	    }
 	}
-	ni.dec_cnt -= zero_cnt;
 	ni.len = pi->cur - ni.str;
     }
+    // Check for special reserved values for Infinity and NaN.
+    if (ni.big) {
+	if (0 == strcasecmp(INF_VAL, ni.str)) {
+	    ni.infinity = 1;
+	} else if (0 == strcasecmp(NINF_VAL, ni.str)) {
+	    ni.infinity = 1;
+	    ni.neg = 1;
+	} else if (0 == strcasecmp(NAN_VAL, ni.str)) {
+	    ni.nan = 1;
+	}
+    }
     if (BigDec == pi->options.bigdec_load) {
 	ni.big = 1;
     }
@@ -517,7 +519,7 @@ read_num(ParseInfo pi) {
 
 static void
 array_start(ParseInfo pi) {
-    VALUE	v = pi->start_array(pi);
+    volatile VALUE	v = pi->start_array(pi);
 
     stack_push(&pi->stack, v, NEXT_ARRAY_NEW);
 }
@@ -547,7 +549,7 @@ static void
 hash_end(ParseInfo pi) {
     volatile Val	hash = stack_peek(&pi->stack);
 
-    // leave hash on stack until just before 
+    // leave hash on stack until just before
     if (0 == hash) {
 	oj_set_error_at(pi, oj_parse_error_class, __FILE__, __LINE__, "unexpected hash close");
     } else if (NEXT_HASH_COMMA != hash->next && NEXT_HASH_NEW != hash->next) {
@@ -587,10 +589,15 @@ colon(ParseInfo pi) {
 
 void
 oj_parse2(ParseInfo pi) {
+    int	first = 1;
+
     pi->cur = pi->json;
     err_init(&pi->err);
     while (1) {
 	next_non_white(pi);
+	if (!first && '\0' != *pi->cur) {
+	    oj_set_error_at(pi, oj_parse_error_class, __FILE__, __LINE__, "unexpected characters after the JSON document");
+	}
 	switch (*pi->cur++) {
 	case '{':
 	    hash_start(pi);
@@ -646,6 +653,9 @@ oj_parse2(ParseInfo pi) {
 	    break;
 	case '/':
 	    skip_comment(pi);
+	    if (first) {
+		continue;
+	    }
 	    break;
 	case '\0':
 	    pi->cur--;
@@ -657,18 +667,23 @@ oj_parse2(ParseInfo pi) {
 	if (err_has(&pi->err)) {
 	    return;
 	}
-	if (Qundef != pi->proc && stack_empty(&pi->stack)) {
-	    if (Qnil == pi->proc) {
-		rb_yield(stack_head_val(&pi->stack));
-	    } else {
+	if (stack_empty(&pi->stack)) {
+	    if (Qundef != pi->proc) {
+		if (Qnil == pi->proc) {
+		    rb_yield(stack_head_val(&pi->stack));
+		} else {
 #if HAS_PROC_WITH_BLOCK
-		VALUE	args[1];
+		    VALUE	args[1];
 
-		*args = stack_head_val(&pi->stack);
-		rb_proc_call_with_block(pi->proc, 1, args, Qnil);
+		    *args = stack_head_val(&pi->stack);
+		    rb_proc_call_with_block(pi->proc, 1, args, Qnil);
 #else
-		rb_raise(rb_eNotImpError, "Calling a Proc with a block not supported in this version. Use func() {|x| } syntax instead.");
+		    rb_raise(rb_eNotImpError,
+			     "Calling a Proc with a block not supported in this version. Use func() {|x| } syntax instead.");
 #endif
+		}
+	    } else {
+		first = 0;
 	    }
 	}
     }
@@ -676,7 +691,7 @@ oj_parse2(ParseInfo pi) {
 
 VALUE
 oj_num_as_value(NumInfo ni) {
-    VALUE	rnum = Qnil;
+    volatile VALUE	rnum = Qnil;
 
     if (ni->infinity) {
 	if (ni->neg) {
@@ -716,15 +731,30 @@ oj_num_as_value(NumInfo ni) {
 		rnum = rb_funcall(rnum, rb_intern("to_f"), 0);
 	    }
 	} else {
-	    double	d = (double)ni->i + (double)ni->num * (1.0L / ni->div);
+	    // All these machinations are to get rounding to work better.
+	    long double	d = (long double)ni->i * (long double)ni->div + (long double)ni->num;
+	    int		x = ni->exp - ni->di;
 
-	    if (ni->neg) {
-		d = -d;
+	    // Rounding sometimes cuts off the last digit even if there are only
+	    // 15 digits. This attempts to fix those few cases where this
+	    // occurs.
+	    if ((long double)INT64_MAX > d && (int64_t)d != (ni->i * ni->div + ni->num)) {
+		rnum = rb_funcall(oj_bigdecimal_class, oj_new_id, 1, rb_str_new(ni->str, ni->len));
+		if (ni->no_big) {
+		    rnum = rb_funcall(rnum, rb_intern("to_f"), 0);
+		}
+	    } else {
+		d = roundl(d);
+		if (0 < x) {
+		    d *= powl(10.0L, x);
+		} else if (0 > x) {
+		    d /= powl(10.0L, -x);
+		}
+		if (ni->neg) {
+		    d = -d;
+		}
+		rnum = rb_float_new((double)d);
 	    }
-	    if (0 != ni->exp) {
-		d *= pow(10.0L, ni->exp);
-	    }
-	    rnum = rb_float_new(d);
 	}
     }
     return rnum;
@@ -753,10 +783,19 @@ protect_parse(VALUE pip) {
     return Qnil;
 }
 
-void
-oj_pi_set_input_str(ParseInfo pi, volatile VALUE input) {
-    pi->json = rb_string_value_ptr((VALUE*)&input);
-    pi->end = pi->json + RSTRING_LEN(input);
+extern int oj_utf8_index;
+
+static void
+oj_pi_set_input_str(ParseInfo pi, volatile VALUE *inputp) {
+#if HAS_ENCODING_SUPPORT
+    rb_encoding	*enc = rb_to_encoding(rb_obj_encoding(*inputp));
+
+    if (rb_utf8_encoding() != enc) {
+	*inputp = rb_str_conv_enc(*inputp, enc, rb_utf8_encoding());
+    }
+#endif
+    pi->json = rb_string_value_ptr((VALUE*)inputp);
+    pi->end = pi->json + RSTRING_LEN(*inputp);
 }
 
 VALUE
@@ -764,7 +803,7 @@ oj_pi_parse(int argc, VALUE *argv, ParseInfo pi, char *json, size_t len, int yie
     char		*buf = 0;
     volatile VALUE	input;
     volatile VALUE	wrapped_stack;
-    VALUE		result = Qnil;
+    volatile VALUE	result = Qnil;
     int			line = 0;
     int			free_json = 0;
 
@@ -785,7 +824,7 @@ oj_pi_parse(int argc, VALUE *argv, ParseInfo pi, char *json, size_t len, int yie
 	pi->end = json + len;
 	free_json = 1;
     } else if (T_STRING == rb_type(input)) {
-	oj_pi_set_input_str(pi, input);
+	oj_pi_set_input_str(pi, &input);
     } else if (Qnil == input && Yes == pi->options.nilnil) {
 	return Qnil;
     } else {
@@ -794,7 +833,7 @@ oj_pi_parse(int argc, VALUE *argv, ParseInfo pi, char *json, size_t len, int yie
 
 	if (oj_stringio_class == clas) {
 	    s = rb_funcall2(input, oj_string_id, 0, 0);
-	    oj_pi_set_input_str(pi, s);
+	    oj_pi_set_input_str(pi, &s);
 #if !IS_WINDOWS
 	} else if (rb_cFile == clas && 0 == FIX2INT(rb_funcall(input, oj_pos_id, 0))) {
 	    int		fd = FIX2INT(rb_funcall(input, oj_fileno_id, 0));
@@ -880,22 +919,35 @@ oj_pi_parse(int argc, VALUE *argv, ParseInfo pi, char *json, size_t len, int yie
 	rb_jump_tag(line);
     }
     if (err_has(&pi->err)) {
+	if (Qnil != pi->err_class) {
+	    pi->err.clas = pi->err_class;
+	}
 	oj_err_raise(&pi->err);
     }
     if (pi->options.quirks_mode == No) {
 	switch (rb_type(result)) {
-            case T_NIL:
-	    case T_TRUE:
-	    case T_FALSE:
-	    case T_FIXNUM:
-	    case T_FLOAT:
-	    case T_CLASS:
-	    case T_SYMBOL:
-		rb_raise(oj_parse_error_class, "unexpected non-document value");
-		break;
-	    default:
-		// okay
-		break;
+	case T_NIL:
+	case T_TRUE:
+	case T_FALSE:
+	case T_FIXNUM:
+	case T_FLOAT:
+	case T_CLASS:
+	case T_STRING:
+	case T_SYMBOL: {
+	    struct _Err	err;
+
+	    if (Qnil == pi->err_class) {
+		err.clas = oj_parse_error_class;
+	    } else {
+		err.clas = pi->err_class;
+	    }
+	    snprintf(err.msg, sizeof(err.msg), "unexpected non-document value");
+	    oj_err_raise(&err);
+	    break;
+	}
+	default:
+	    // okay
+	    break;
 	}
     }
     return result;
